/*******************************************************************************

  Copyright (c) 2001-2003 Intel Corporation All rights reserved. 
  
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions 
  are met: 
  
  - Redistributions of source code must retain the above copyright 
  notice, this list of conditions and the following disclaimer. 
  - Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution. 
  - Neither the name of the Intel Corporation nor the names of its 
  contributors may be used to endorse or promote products derived from 
  this software without specific prior written permission.
   
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL 
  OR ITS  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/


/*
 * Proc fs support.
 *
 * Read-only files created by driver (if CONFIG_PROC_FS):
 *
 * /proc/net/PRO_LAN_Adapters/<ethx>.info
 * /proc/net/PRO_LAN_Adapters/<ethx>/<attribute>
 *
 * where <ethx>       is the system device name, i.e eth0.
 *       <attribute>  is the driver attribute name.
 *
 * There is one file for each driver attribute, where the contents
 * of the file is the attribute value.  The ethx.info file contains
 * a list of all driver attributes in one file.
 *
 */

#include "ixgb.h"

#ifdef CONFIG_PROC_FS

#include <linux/proc_fs.h>

#define ADAPTERS_PROC_DIR           "PRO_LAN_Adapters"
#define TAG_MAX_LENGTH              32
#define LINE_MAX_LENGTH             80
#define FIELD_MAX_LENGTH            LINE_MAX_LENGTH - TAG_MAX_LENGTH - 3

extern char ixgb_driver_name[];
extern char ixgb_driver_version[];

/*
 * The list of driver proc attributes is stored in a proc_list link
 * list.  The list is build with proc_list_setup and is used to
 * build the proc fs nodes.  The private data for each node is the
 * corresponding link in the link list.
 */

struct proc_list {
        struct list_head list;    /* link list        */
        char tag[TAG_MAX_LENGTH + 1];           /* attribute name */
        void *data;               /* attribute data   */
        size_t len;               /* sizeof data      */
        char *(*func)(void *, size_t, char *);  /* format data func */
};

static int
ixgb_proc_read(char *page, char **start, off_t off, int count, int *eof)
{
        int len = strlen(page);

        page[len++] = '\n';

        if(len <= off + count)
                *eof = 1;
        *start = page + off;
        len -= off;
        if(len > count)
                len = count;
        if(len < 0)
                len = 0;

        return len;
}

static int
ixgb_proc_info_read(char *page, char **start, off_t off,
                    int count, int *eof, void *data)
{
        struct list_head *proc_list_head = data, *curr;
        struct proc_list *elem;
        char *p = page;
        char buf[FIELD_MAX_LENGTH + 1];

        list_for_each(curr, proc_list_head) {
                elem = list_entry(curr, struct proc_list, list);

                if (p - page + LINE_MAX_LENGTH >= PAGE_SIZE)
                   break;

                if(!strlen(elem->tag))
                        p += sprintf(p, "\n");
                else
                        p += sprintf(p, "%-*.*s %.*s\n", 
                                     TAG_MAX_LENGTH, TAG_MAX_LENGTH,
                                     elem->tag, FIELD_MAX_LENGTH,
                                     elem->func(elem->data, elem->len, buf));
        }

        *p = '\0';

        return ixgb_proc_read(page, start, off, count, eof);
}

static int
ixgb_proc_single_read(char *page, char **start, off_t off,
                      int count, int *eof, void *data)
{
        struct proc_list *elem = data;

        sprintf(page, "%.*s", FIELD_MAX_LENGTH, elem->func(elem->data, 
                elem->len, page));

        return ixgb_proc_read(page, start, off, count, eof);
}

static void __devexit
ixgb_proc_dirs_free(char *name, struct list_head *proc_list_head)
{
        struct proc_dir_entry *intel_proc_dir, *proc_dir;
        char info_name[strlen(name) + strlen(".info")];

        for(intel_proc_dir = proc_net->subdir; intel_proc_dir;
            intel_proc_dir = intel_proc_dir->next) {
                if((intel_proc_dir->namelen == strlen(ADAPTERS_PROC_DIR)) &&
                   !memcmp(intel_proc_dir->name, 
                           ADAPTERS_PROC_DIR, 
                           strlen(ADAPTERS_PROC_DIR)))
                        break;
        }

        if(!intel_proc_dir)
                return;

        for(proc_dir = intel_proc_dir->subdir; proc_dir;
            proc_dir = proc_dir->next) {
                if((proc_dir->namelen == strlen(name)) &&
                   !memcmp(proc_dir->name, name, strlen(name)))
                        break;
        }

        if(proc_dir) {
                struct list_head *curr;
                struct proc_list *elem;

                list_for_each(curr, proc_list_head) {
                        elem = list_entry(curr, struct proc_list, list);
                        remove_proc_entry(elem->tag, proc_dir);
                } 

                strcpy(info_name, name);
                strcat(info_name, ".info");

                remove_proc_entry(info_name, intel_proc_dir);
                remove_proc_entry(name, intel_proc_dir);
        }

        /* If the intel dir is empty, remove it */

        for(proc_dir = intel_proc_dir->subdir; proc_dir;
            proc_dir = proc_dir->next) {

                /* ignore . and .. */

                if(*(proc_dir->name) == '.')
                        continue;
                break;
        }

        if(!proc_dir)
                remove_proc_entry(ADAPTERS_PROC_DIR, proc_net);
}


static int __devinit
ixgb_proc_singles_create(struct proc_dir_entry *parent,
                         struct list_head *proc_list_head)
{
        struct list_head *curr;
        struct proc_list *elem;

        list_for_each(curr, proc_list_head) {
                struct proc_dir_entry *proc_entry;

                elem = list_entry(curr, struct proc_list, list);

                if(!strlen(elem->tag))
                        continue;

                if(!(proc_entry = 
                        create_proc_entry(elem->tag, S_IFREG, parent)))
                        return 0;

                proc_entry->read_proc = ixgb_proc_single_read;
                proc_entry->data = elem;
                SET_MODULE_OWNER(proc_entry);
        }

        return 1;
}

static void __devinit
ixgb_proc_dirs_create(char *name, struct list_head *proc_list_head)
{
        struct proc_dir_entry *intel_proc_dir, *proc_dir, *info_entry;
        char info_name[strlen(name) + strlen(".info")];

        for(intel_proc_dir = proc_net->subdir; intel_proc_dir;
            intel_proc_dir = intel_proc_dir->next) {
                if((intel_proc_dir->namelen == strlen(ADAPTERS_PROC_DIR)) &&
                   !memcmp(intel_proc_dir->name, ADAPTERS_PROC_DIR,
                          strlen(ADAPTERS_PROC_DIR)))
                        break;
        }

        if(!intel_proc_dir)
                if(!(intel_proc_dir =
                        create_proc_entry(ADAPTERS_PROC_DIR,
                                S_IFDIR, proc_net)))
                        return;

        if(!(proc_dir =
                create_proc_entry(name, S_IFDIR, intel_proc_dir)))
                return;
        SET_MODULE_OWNER(proc_dir);

        if(!ixgb_proc_singles_create(proc_dir, proc_list_head))
                return;

        strcpy(info_name, name);
        strcat(info_name, ".info");

        if(!(info_entry = 
                create_proc_entry(info_name, S_IFREG, intel_proc_dir)))
                return;
        SET_MODULE_OWNER(info_entry);

        info_entry->read_proc = ixgb_proc_info_read;
        info_entry->data = proc_list_head;
}

static void __devinit
ixgb_proc_list_add(struct list_head *proc_list_head, char *tag,
                   void *data, size_t len,
                   char *(*func) (void *, size_t, char *))
{
        struct proc_list *new = (struct proc_list *) 
                kmalloc(sizeof(struct proc_list), GFP_KERNEL);

        if(!new)
                return;

        strncpy(new->tag, tag, TAG_MAX_LENGTH);
        new->data = data;
        new->len = len;
        new->func = func;

        list_add_tail(&new->list, proc_list_head);
}

static void __devexit
ixgb_proc_list_free(struct list_head *proc_list_head)
{
        struct proc_list *elem;

        while(!list_empty(proc_list_head)) {
                elem = list_entry(proc_list_head->next, struct proc_list, list);
                list_del(&elem->list);
                kfree(elem);
        }
}

/*
 * General purpose formating functions
 */

static char *
ixgb_proc_str(void *data, size_t len, char *buf)
{
        sprintf(buf, "%s", (char *) data);
        return buf;
}

static char *
ixgb_proc_hex(void *data, size_t len, char *buf)
{
        switch (len) {
        case sizeof(uint8_t):
                sprintf(buf, "0x%02x", *(uint8_t *) data);
                break;
        case sizeof(uint16_t):
                sprintf(buf, "0x%04x", *(uint16_t *) data);
                break;
        case sizeof(uint32_t):
                sprintf(buf, "0x%08x", *(uint32_t *) data);
                break;
        case sizeof(uint64_t):
                sprintf(buf, "0x%08Lx", (unsigned long long)*(uint64_t *)data);
                break;
        }
        return buf;
}

static char *
ixgb_proc_unsigned(void *data, size_t len, char *buf)
{
        switch (len) {
        case sizeof(uint8_t):
                sprintf(buf, "%u", *(uint8_t *) data);
                break;
        case sizeof(uint16_t):
                sprintf(buf, "%u", *(uint16_t *) data);
                break;
        case sizeof(uint32_t):
                sprintf(buf, "%u", *(uint32_t *) data);
                break;
        case sizeof(uint64_t):
                sprintf(buf, "%Lu", (unsigned long long) *(uint64_t *) data);
                break;
        }
        return buf;
}

/*
 * Specific formating functions
 */

static char *
ixgb_proc_part_number(void *data, size_t len, char *buf)
{
        uint32_t d = *(uint32_t *)data;

        /* swap words in d */
        d = (((d & 0xFFFF0000) >> 16) | ((d & 0x0000FFFF) << 16));

        sprintf(buf, "%06x-%03x", d >> 8, d & 0x000000FF);
        return buf;
}

static char *
ixgb_proc_slot(void *data, size_t len, char *buf)
{
        struct ixgb_adapter *adapter = data;
        sprintf(buf, "%u", PCI_SLOT(adapter->pdev->devfn));
        return buf;
}

static char *
ixgb_proc_bus_type(void *data, size_t len, char *buf)
{
        ixgb_bus_type bus_type = *(ixgb_bus_type *) data;
        sprintf(buf,
                bus_type == ixgb_bus_type_pci ? "PCI" : 
                bus_type == ixgb_bus_type_pcix ? "PCI-X" : 
                "UNKNOWN");
        return buf;
}

static char *
ixgb_proc_bus_speed(void *data, size_t len, char *buf)
{
        ixgb_bus_speed bus_speed = *(ixgb_bus_speed *) data;
        sprintf(buf,
                bus_speed == ixgb_bus_speed_33 ? "33MHz" : 
                bus_speed == ixgb_bus_speed_66 ? "66MHz" : 
                bus_speed == ixgb_bus_speed_100 ? "100MHz" : 
                bus_speed == ixgb_bus_speed_133 ? "133MHz" : 
                "UNKNOWN");
        return buf;
}

static char *
ixgb_proc_bus_width(void *data, size_t len, char *buf)
{
        ixgb_bus_width bus_width = *(ixgb_bus_width *) data;
        sprintf(buf,
                bus_width == ixgb_bus_width_32 ? "32-bit" : 
                bus_width == ixgb_bus_width_64 ? "64-bit" : 
                "UNKNOWN");
        return buf;
}

static char *
ixgb_proc_hwaddr(void *data, size_t len, char *buf)
{
        unsigned char *hwaddr = data;
        sprintf(buf, "%02X:%02X:%02X:%02X:%02X:%02X", 
                hwaddr[0], hwaddr[1], hwaddr[2], 
                hwaddr[3], hwaddr[4], hwaddr[5]);
        return buf;
}

static char *
ixgb_proc_link(void *data, size_t len, char *buf)
{
        struct ixgb_adapter *adapter = data;
        sprintf(buf, netif_running(adapter->netdev) ? 
                netif_carrier_ok(adapter->netdev) ? 
                "up" : "down" : "N/A");
        return buf;
}

static char *
ixgb_proc_link_speed(void *data, size_t len, char *buf)
{
        struct ixgb_adapter *adapter = data;
        sprintf(buf, adapter->link_speed ? "%u" : "N/A", adapter->link_speed);
        return buf;
}

static char *
ixgb_proc_link_duplex(void *data, size_t len, char *buf)
{
        struct ixgb_adapter *adapter = data;
        sprintf(buf, adapter->link_duplex == FULL_DUPLEX ? "Full" : "N/A");
        return buf;
}

static char *
ixgb_proc_state(void *data, size_t len, char *buf)
{
        struct ixgb_adapter *adapter = data;
        sprintf(buf, adapter->netdev->flags & IFF_UP ? "up" : "down");
        return buf;
}

/*
 * ixgb_proc_list_setup - build link list of proc praramters
 * @adapter: board private structure
 *
 * Order matters - ethx.info entries are ordered in the order links 
 * are added to list.
 */

#define LIST_ADD_F(T,D,F) \
    ixgb_proc_list_add(proc_list_head, (T), (D), sizeof(*(D)), (F))
#define LIST_ADD_BLANK() LIST_ADD_F("", NULL, NULL)
#define LIST_ADD_S(T,D) LIST_ADD_F((T), (D), ixgb_proc_str)
#define LIST_ADD_H(T,D) LIST_ADD_F((T), (D), ixgb_proc_hex)
#define LIST_ADD_U(T,D) LIST_ADD_F((T), (D), ixgb_proc_unsigned)

static void __devinit
ixgb_proc_list_setup(struct ixgb_adapter *adapter)
{
        struct ixgb_hw *hw = &adapter->hw;
        struct list_head *proc_list_head = &adapter->proc_list_head;

        INIT_LIST_HEAD(proc_list_head);
        LIST_ADD_S("Description", adapter->id_string);
        LIST_ADD_F("Part_Number", &adapter->part_num, ixgb_proc_part_number);
        LIST_ADD_S("Driver_Name", ixgb_driver_name);
        LIST_ADD_S("Driver_Version", ixgb_driver_version);
        LIST_ADD_H("PCI_Vendor", &hw->vendor_id);
        LIST_ADD_H("PCI_Device_ID", &hw->device_id);
        LIST_ADD_H("PCI_Subsystem_Vendor", &hw->subsystem_vendor_id);
        LIST_ADD_H("PCI_Subsystem_ID", &hw->subsystem_id);
        LIST_ADD_H("PCI_Revision_ID", &hw->revision_id);
        LIST_ADD_U("PCI_Bus", &adapter->pdev->bus->number);
        LIST_ADD_F("PCI_Slot", adapter, ixgb_proc_slot);
        LIST_ADD_F("PCI_Bus_Type", &hw->bus.type, ixgb_proc_bus_type);
        LIST_ADD_F("PCI_Bus_Speed", &hw->bus.speed, ixgb_proc_bus_speed);
        LIST_ADD_F("PCI_Bus_Width", &hw->bus.width, ixgb_proc_bus_width);
        LIST_ADD_U("IRQ", &adapter->pdev->irq);
        LIST_ADD_S("System_Device_Name", adapter->ifname);
        LIST_ADD_F("Current_HWaddr", adapter->netdev->dev_addr,
                   ixgb_proc_hwaddr);
        LIST_ADD_F("Permanent_HWaddr", adapter->hw.eeprom, ixgb_proc_hwaddr);
        LIST_ADD_BLANK();

        LIST_ADD_F("Link", adapter, ixgb_proc_link);
        LIST_ADD_F("Speed", adapter, ixgb_proc_link_speed);
        LIST_ADD_F("Duplex", adapter, ixgb_proc_link_duplex);
        LIST_ADD_F("State", adapter, ixgb_proc_state);
        LIST_ADD_BLANK();

        /* Standard net device stats */
        LIST_ADD_U("Rx_Packets", &adapter->net_stats.rx_packets);
        LIST_ADD_U("Tx_Packets", &adapter->net_stats.tx_packets);
        LIST_ADD_U("Rx_Bytes", &adapter->net_stats.rx_bytes);
        LIST_ADD_U("Tx_Bytes", &adapter->net_stats.tx_bytes);
        LIST_ADD_U("Rx_Errors", &adapter->net_stats.rx_errors);
        LIST_ADD_U("Tx_Errors", &adapter->net_stats.tx_errors);
        LIST_ADD_U("Rx_Dropped", &adapter->net_stats.rx_dropped);
        LIST_ADD_U("Tx_Dropped", &adapter->net_stats.tx_dropped);

        LIST_ADD_U("Multicast", &adapter->net_stats.multicast);
        LIST_ADD_U("Collisions", &adapter->net_stats.collisions);

        /* LIST_ADD_U("Rx_Length_Errors", &adapter->net_stats.rx_length_errors);
         */
        LIST_ADD_U("Rx_Over_Errors", &adapter->net_stats.rx_over_errors);
        LIST_ADD_U("Rx_CRC_Errors", &adapter->net_stats.rx_crc_errors);
        LIST_ADD_U("Rx_Frame_Errors", &adapter->net_stats.rx_frame_errors);
        LIST_ADD_U("Rx_FIFO_Errors", &adapter->net_stats.rx_fifo_errors);
        LIST_ADD_U("Rx_Missed_Errors", &adapter->net_stats.rx_missed_errors);

        LIST_ADD_U("Tx_Aborted_Errors", &adapter->net_stats.tx_aborted_errors);
        LIST_ADD_U("Tx_Carrier_Errors", &adapter->net_stats.tx_carrier_errors);
        LIST_ADD_U("Tx_FIFO_Errors", &adapter->net_stats.tx_fifo_errors);
        LIST_ADD_U("Tx_Heartbeat_Errors",
                    &adapter->net_stats.tx_heartbeat_errors);
        LIST_ADD_U("Tx_Window_Errors", &adapter->net_stats.tx_window_errors);

        /* 82597-specific stats */
        LIST_ADD_U("Tx_Deferred_Ok", &adapter->stats.dc);
        LIST_ADD_U("Rx_Long_Length_Errors", &adapter->stats.roc);
        LIST_ADD_U("Rx_Short_Length_Errors", &adapter->stats.ruc);

        LIST_ADD_U("Rx_Flow_Control_XON", &adapter->stats.xonrxc);
        LIST_ADD_U("Rx_Flow_Control_XOFF", &adapter->stats.xoffrxc);
        LIST_ADD_U("Tx_Flow_Control_XON", &adapter->stats.xontxc);
        LIST_ADD_U("Tx_Flow_Control_XOFF", &adapter->stats.xofftxc);
        LIST_ADD_U("Rx_CSum_Offload_Good", &adapter->hw_csum_rx_good);
        LIST_ADD_U("Rx_CSum_Offload_Errors", &adapter->hw_csum_rx_error);
        LIST_ADD_U("Tx_CSum_Offload_Good", &adapter->hw_csum_tx_good);
        LIST_ADD_U("Tx_CSum_Offload_Errors", &adapter->hw_csum_tx_error);
#ifdef NETIF_F_TSO
        LIST_ADD_U("TCP Seg Context Transmitted", &adapter->stats.tsctc);
        LIST_ADD_U("TCP Seg Context Tx Fail", &adapter->stats.tsctfc);
#endif
        LIST_ADD_BLANK();
}

/*
 * ixgb_proc_dev_setup - create proc fs nodes and link list
 * @adapter: board private structure
 */
void __devinit
ixgb_proc_dev_setup(struct ixgb_adapter *adapter)
{
        ixgb_proc_list_setup(adapter);
        ixgb_proc_dirs_create(adapter->ifname, &adapter->proc_list_head);
}

/*
 * ixgb_proc_dev_free - free proc fs nodes and link list
 * @adapter: board private structure
 */
void __devexit
ixgb_proc_dev_free(struct ixgb_adapter *adapter)
{
        ixgb_proc_dirs_free(adapter->ifname, &adapter->proc_list_head);
        ixgb_proc_list_free(&adapter->proc_list_head);
}
#else /* CONFIG_PROC_FS */

void __devinit ixgb_proc_dev_setup(struct ixgb_adapter *adapter) {} 
void __devexit ixgb_proc_dev_free(struct ixgb_adapter *adapter) {}

#endif /* CONFIG_PROC_FS */
