/*******************************************************************************

  Copyright (c) 2001-2003 Intel Corporation All rights reserved. 
  
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions 
  are met: 
  
  - Redistributions of source code must retain the above copyright 
  notice, this list of conditions and the following disclaimer. 
  - Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution. 
  - Neither the name of the Intel Corporation nor the names of its 
  contributors may be used to endorse or promote products derived from 
  this software without specific prior written permission.
   
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL 
  OR ITS  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include "kcompat.h"

/*****************************************************************************/

/* 2.4.13 => 2.4.3 */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,13) )

/**************************************/

/* PCI DMA MAPPING */

#if defined(CONFIG_HIGHMEM)

#ifndef PCI_DRAM_OFFSET
#define PCI_DRAM_OFFSET 0
#endif

u64
_kc_pci_map_page(struct pci_dev *dev,
                 struct page *page,
                 unsigned long offset,
                 size_t size,
                 int direction)
{
        return (((u64) (page - mem_map) << PAGE_SHIFT) + offset +
                PCI_DRAM_OFFSET);
}

#else /* CONFIG_HIGHMEM */

u64
_kc_pci_map_page(struct pci_dev *dev,
                 struct page *page,
                 unsigned long offset,
                 size_t size,
                 int direction)
{
        return pci_map_single(dev, (void *) page_address(page) + offset, size,
                              direction);
}

#endif /* CONFIG_HIGHMEM */

void
_kc_pci_unmap_page(struct pci_dev *dev,
                   u64 dma_addr,
                   size_t size,
                   int direction)
{
        return pci_unmap_single(dev, dma_addr, size, direction);
}

#endif /* 2.4.13 => 2.4.3 */

/*****************************************************************************/

/* 2.4.3 => 2.4.0 */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,3) )

/**************************************/

/* PCI DRIVER API */

_kc_pci_set_dma_mask(struct pci_dev * dev, dma_addr_t mask)
{
        if(!pci_dma_supported(dev, mask))
                return -EIO;
        dev->dma_mask = mask;
        return 0;
}

int
_kc_pci_request_regions(struct pci_dev *dev,
                        char *res_name)
{
        int i;

        for(i = 0; i < 6; i++) {
                if(pci_resource_len(dev, i) == 0)
                        continue;

                if(pci_resource_flags(dev, i) & IORESOURCE_IO) {
                        if(!request_region
                           (pci_resource_start(dev, i),
                            pci_resource_len(dev, i), res_name)) {
                                pci_release_regions(dev);
                                return -EBUSY;
                        }
                } else if(pci_resource_flags(dev, i) & IORESOURCE_MEM) {
                        if(!request_mem_region
                           (pci_resource_start(dev, i),
                            pci_resource_len(dev, i), res_name)) {
                                pci_release_regions(dev);
                                return -EBUSY;
                        }
                }
        }
        return 0;
}

void
_kc_pci_release_regions(struct pci_dev *dev)
{
        int i;

        for(i = 0; i < 6; i++) {
                if(pci_resource_len(dev, i) == 0)
                        continue;

                if(pci_resource_flags(dev, i) & IORESOURCE_IO)
                        release_region(pci_resource_start(dev, i),
                                       pci_resource_len(dev, i));

                else if(pci_resource_flags(dev, i) & IORESOURCE_MEM)
                        release_mem_region(pci_resource_start(dev, i),
                                           pci_resource_len(dev, i));
        }
}

/**************************************/

/* NETWORK DRIVER API */

struct net_device *
_kc_alloc_etherdev(int sizeof_priv)
{
        struct net_device *dev;
        int alloc_size;

        alloc_size = sizeof(*dev) + sizeof_priv + IFNAMSIZ + 31;

        dev = kmalloc(alloc_size, GFP_KERNEL);

        if(!dev)
                return NULL;

        memset(dev, 0, alloc_size);

        if(sizeof_priv)
                dev->priv = (void *) (((unsigned long) (dev + 1) + 31) & ~31);

        dev->name[0] = '\0';

        ether_setup(dev);

        return dev;
}

int
_kc_is_valid_ether_addr(u8 * addr)
{
        const char zaddr[6] = { 0, };

        return !(addr[0] & 1) && memcmp(addr, zaddr, 6);
}

#endif /* 2.4.3 => 2.4.0 */
